//
// X11 Display abstraction classes for PTC 2.0 C++ API
// Copyright (c) 1998 Christian Nentwich (brn@eleet.mcb.at)
// The PTC 2.0 C++ API is (c) 1998 Glenn Fiedler (ptc@gaffer.org)
// This source code is licensed under the GNU LGPL
//
// Please refer to the file COPYING.LIB contained in the distribution for
// licensing conditions
//

#ifndef __PTC_X11DISPLAY_H
#define __PTC_X11DISPLAY_H

#include "ptconfig.h"
#include "Format.h"
#include "Copy.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "X11Image.h"

class X11Display {

public:
  X11Display();
  virtual ~X11Display();

  static bool checkDGA(int width,int height,Display *disp,int screen);
  
  virtual void open(const char title[],int width,int height,Format& format,
		    Display *disp,int screen) = 0;

  // This will return a windowed object
  virtual void open(Display *disp,int screen,Window w,Format& format) = 0;

  virtual void close() = 0;

  // Set the X11 specific flags
  void flags(long mask) { m_flags=mask; }
  void addflags(long mask) { m_flags|=mask; }

  virtual void update() = 0;

  virtual void write(void *pixels,int width,int height,Format& format, 
		     int32 palette[]);
  
  virtual bool key();
  virtual int read();

  virtual void* lock() = 0;
  virtual void unlock() = 0;

  virtual void palette(int32 palette[]) = 0;

  int width();
  int height();
  const Format& format();

  virtual int pitch() = 0;


  // X11 helper functions for your enjoyment
  Display* getDisplay() const { return m_disp; }
  int getScreen() const { return m_screen; }


protected:
  // Generate internal format from requested format
  Format getFormat(Format &format);          

  long m_flags;
  unsigned int m_width;
  unsigned int m_height;
  Format m_format;

  Copy m_copy;

  Display *m_disp;                                // X11 Display
  int m_screen;                                   // X11 Screen

  Colormap m_cmap;                                // Color map
  XColor *m_colours;                              // Color cells
};


class X11WindowDisplay : public X11Display {
public:
  X11WindowDisplay();
  ~X11WindowDisplay();

  virtual void open(const char title[],int width,int height,Format& format,
		    Display *disp,int screen);
  virtual void open(Display *disp,int screen,Window w,Format& format);

  virtual void close();

  virtual void update();

  virtual void* lock();
  virtual void unlock();

  virtual void palette(int32 palette[]);

  virtual int pitch();


  // X11 convenience functions 
  Window getWindow() { return m_window; }
  GC getGC() { return m_gc; }

private:
  X11Image *createImage(Display *disp,int screen,int width,int height,
			Format &format);          // Factory method

  Window m_window;
  Atom m_atom_close;

  X11Image *m_primary;
  GC m_gc;
};

#ifdef HAVE_DGA

#include <X11/extensions/xf86vmode.h>

class X11DGADisplay : public X11Display {
public:
  ~X11DGADisplay();

  virtual void open(const char title[],int width,int height,Format& format,
		    Display *disp,int screen);
  // Not in DGA mode
  virtual void open(Display *disp,int screen,Window w,Format& format) {}

  virtual void close();

  virtual void update();

  virtual void* lock();
  virtual void unlock();

  virtual void palette(int32 palette[]);

  virtual int pitch();

private:
  XF86VidModeModeInfo **modeinfo;
  int num_modeinfo;
  int previousmode;
  
  char *dga_addr;
  int dga_linewidth;
  int dga_banksize;
  int dga_memsize;
};

#else

class X11DGADisplay : public X11Display {
};

#endif

#endif



